///////////////////////////////////////////////////////////////////////////////
// Projet :  Recherche de solution du Solitaire
//           Je donne mon accord  developpez.com pour une ventuelle
//           intgration dans les pages Sources du site.
// Fichier : Solveur.cpp
// Auteur :  Sylvain Togni
// Date :    08/03/2009
///////////////////////////////////////////////////////////////////////////////
#include "Solveur.h"
#include "Plateau.h"
#include <set>
#include <vector>
#include <algorithm>
#include <utility>

///////////////////////////////////////////////////////////////////////////////
// Fonction pour vrifier la non-faisabilit d'un plateau (rgle de trois).
// http://eternitygames.free.fr/Solitaire.html#La%20r%C3%A8gle%20de%20trois
///////////////////////////////////////////////////////////////////////////////
static std::pair<int, int> phi(Plateau const& p)
{
    int n1[3] = {0};
    int n2[3] = {0};
    for(std::size_t i = 0; i < p.hauteur(); ++i)
    {
        for(std::size_t j = 0; j < p.largeur(); ++j)
        {
            if (p(i, j) == 'x')
            {
                ++n1[(i + j)%3];
                ++n2[(i + p.largeur()*3 - j)%3];
            }
        }
    }
    static int const classes[2][2][2] = {{{0, 3}, {2, 1}}, {{1, 2}, {3, 0}}};
    return std::pair<int, int>(
        classes[n1[0]%2][n1[1]%2][n1[2]%2],
        classes[n2[0]%2][n2[1]%2][n2[2]%2]);
}

///////////////////////////////////////////////////////////////////////////////
// Heuristique pour juger de la pertinance d'un plateau.
///////////////////////////////////////////////////////////////////////////////
static std::size_t heuristique(Plateau const& p)
{
    // Calcul de la somme des carrs des carts types des coordonnes des pions
    // Plus cette valeur est faible, plus le plateau  de chance de mener  une
    // solution.
    std::size_t s1 = 0;
    std::size_t si = 0;
    std::size_t sj = 0;
    std::size_t sij2 = 0;
    for(std::size_t i = 0; i < p.hauteur(); ++i)
    {
        for(std::size_t j = 0; j < p.largeur(); ++j)
        {
            if (p(i, j) == 'x')
            {
                s1 += 1;
                si += i;
                sj += j;
                sij2 += i*i + j*j;
            }
        }
    }
    return sij2*s1 - si*si - sj*sj;
}

//////////////////////////////////////////////////////////////////////////////
// Structure reprsentant un noeud dans le graphe des positions.
//////////////////////////////////////////////////////////////////////////////
struct Noeud
{
    // Valeur de l'heuristique
    std::size_t heuristique;

    // Plateau correspondant
    Plateau plateau;

    // Liste des coups menant  ce noeud
    // NB : un vector<char> semble plus performant qu'une std::string ici
    std::vector<char> chemin;

    // Constructeur
    Noeud(std::size_t h, Plateau const& p) : heuristique(h), plateau(p) {}

    // Relation d'ordre
    bool operator<(Noeud const& autre) const
    {
	    // Ordonne suivant l'heuristique puis suivant le plateau afin que deux
		// plateaux ayant la mme valeur d'heuristique mais tant diffrents ne
		// soient pas considrs gaux.
        return heuristique < autre.heuristique
            || (heuristique == autre.heuristique && plateau < autre.plateau);
    }

    std::size_t memoireAllouee() const
    {
        return plateau.memoireAllouee() + chemin.capacity();
    }
};

//////////////////////////////////////////////////////////////////////////////
// Un beam est un conteneur tri qui ne garde que les n-meilleurs elements.
//////////////////////////////////////////////////////////////////////////////
class Beam
{
public:
    // Construit un beam d'une certaine taille
    explicit Beam(std::size_t taille)
        : maTaille(taille)
        , monADeborde(false)
    {
    }

    // Insert un element dans le beam
    void insert(Noeud const& n)
    {
        if (size() >= maTaille)
        {
            // beam plein
            std::set<Noeud>::iterator dernier = mesNoeuds.end();
            --dernier;
            if (n < *dernier)
            {
                if (mesNoeuds.insert(n).second)
                {
                    // l'insertion a eut lieu => il faut supprimer le dernier
                    mesNoeuds.erase(dernier);
                    monADeborde = true;
                }
            }
            else
            {
                monADeborde = true;
            }
        }
        else
        {
            // beam non-plein
            mesNoeuds.insert(n);
        }
    }

    // Est-ce que le beam a refus un element ?
    bool aDeborde() const {return monADeborde;}

    // Interface  la STL
    typedef std::set<Noeud>::const_iterator const_iterator;
    const_iterator begin() const {return mesNoeuds.begin();}
    const_iterator end() const {return mesNoeuds.end();}
    std::size_t size() const {return mesNoeuds.size();}
    void swap(Beam& autre) {mesNoeuds.swap(autre.mesNoeuds);}

    std::size_t memoireAllouee() const
    {
        std::size_t resultat = 0;
        for(const_iterator it = begin(); it != end(); ++it)
        {
            resultat += it->memoireAllouee() + sizeof(*it);
            // Pas tout--fait, mais bon...
        }
        return resultat;
    }

private:
    std::set<Noeud>     mesNoeuds;
    std::size_t         maTaille;
    bool                monADeborde;
};

//////////////////////////////////////////////////////////////////////////////
// Teste la validit d'un coup et avance dans le graphe des positions.
//////////////////////////////////////////////////////////////////////////////
static void avance(Beam& noeudsSuivants,
                   Noeud& noeud,   // rutilisation du mme objet pour viter
                                   // des allocations dynamiques inutiles
                   Noeud const& noeudPrecedent,
                   std::size_t i0, std::size_t j0, // coordonnes de dpart
                   std::size_t i1, std::size_t j1, // coordonnes du milieu
                   std::size_t i2, std::size_t j2) // coordonnes d'arrive
{
    if (   noeudPrecedent.plateau(i0, j0) == 'x'
        && noeudPrecedent.plateau(i2, j2) == '.')
    {
        noeud = noeudPrecedent;

        // mise  jour du plateau
        noeud.plateau(i0, j0) = '.';
        noeud.plateau(i1, j1) = '.';
        noeud.plateau(i2, j2) = 'x';

        // calcule de l'heuritique
        noeud.heuristique = heuristique(noeud.plateau);

        // mise  jour du chemin
        noeud.chemin.push_back(static_cast<char>('a' + j0));
        noeud.chemin.push_back(static_cast<char>('1' + i0));
        noeud.chemin.push_back(static_cast<char>('a' + j2));
        noeud.chemin.push_back(static_cast<char>('1' + i2));
        noeud.chemin.push_back(' ');

        // ajout du nouveau noeud
        noeudsSuivants.insert(noeud);
    }
}

//////////////////////////////////////////////////////////////////////////////
// Constructeur
//////////////////////////////////////////////////////////////////////////////
Solveur::Solveur()
: monNombreDePositions(0)
, maMemoireUtilisee(0)
{
}

//////////////////////////////////////////////////////////////////////////////
// Fonction de recherche
//////////////////////////////////////////////////////////////////////////////
bool Solveur::rechercheChemin(Plateau const& p1, Plateau const& p2)
{
    // Initialisation
    maSolution.clear();
    monNombreDePositions = 0;
    maMemoireUtilisee = 0;

    if (phi(p1) == phi(p2)) // critre de non-faisabilit
    {
        // Beam search avec une taille de beam de plus en plus grande
        for(std::size_t taille = 1; beamSearch(p1, p2, taille); taille *= 2)
        {
        }
    }
    return !maSolution.empty();
}

//////////////////////////////////////////////////////////////////////////////
// Algorithme de beam search
//////////////////////////////////////////////////////////////////////////////
bool Solveur::beamSearch(Plateau const& p1, Plateau const& p2,
                         std::size_t taille)
{
    bool encore = false;

    // Initialisation du beam avec la position initiale (p1)
    Beam noeudsPrecedents(taille);
    Noeud noeud(heuristique(p1), p1);
    noeudsPrecedents.insert(noeud);

    // Recherche d'un chemin entre p1 (n1 pions) et p2 (n2 pions)
    std::size_t n1 = p1.nombreDePions();
    std::size_t n2 = p2.nombreDePions();
    for(std::size_t n = n1; n > n2; --n)
    {
        Beam noeudsSuivants(taille);

        // Parcours des noeuds prcdents
        for(Beam::const_iterator it = noeudsPrecedents.begin();
            it != noeudsPrecedents.end(); ++it)
        {
            // Recherche des coups valides
            for(std::size_t i = 0; i < it->plateau.hauteur(); ++i)
            {
                for(std::size_t j = 0; j < it->plateau.largeur(); ++j)
                {
                    if (it->plateau(i, j) == 'x')
                    {
                        // Coups horizontaux
                        if (j > 0 && j + 1 < it->plateau.largeur())
                        {
                            avance(noeudsSuivants, noeud, *it,
                                   i, j - 1, i, j, i, j + 1);
                            avance(noeudsSuivants, noeud, *it,
                                   i, j + 1, i, j, i, j - 1);
                        }
                        // Coups verticaux
                        if (i > 0 && i + 1 < it->plateau.hauteur())
                        {
                            avance(noeudsSuivants, noeud, *it,
                                   i - 1, j, i, j, i + 1, j);
                            avance(noeudsSuivants, noeud, *it,
                                   i + 1, j, i, j, i - 1, j);
                        }
                    }
                }
            }
        }
        if (noeudsSuivants.aDeborde())
        {
            // Prviens l'appelant que toutes les positions n'ont pas t
            // explores
            encore = true;
        }

        // Prpare l'itration suivante
        noeudsSuivants.swap(noeudsPrecedents);
        monNombreDePositions += noeudsPrecedents.size();
        maMemoireUtilisee = std::max(maMemoireUtilisee,
            noeudsSuivants.memoireAllouee()
            + noeudsPrecedents.memoireAllouee());
    }

    // Recherche la position finale
    for(Beam::const_iterator it = noeudsPrecedents.begin();
        it != noeudsPrecedents.end(); ++it)
    {
        if (it->plateau == p2)
        {
            // Solution trouve !
            maSolution.assign(it->chemin.begin(), it->chemin.end());
            encore = false;
        }
    }
    return encore;
}
